#!/usr/bin/env python3

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import io
import os
import platform
import re
import sys
import zipfile
import tarfile
import urllib.request
import shutil
from setuptools import Distribution, setup
from setuptools.command.build_py import build_py
from setuptools.command.bdist_wheel import bdist_wheel

CREDENTIAL_PROVIDER_BASE = "https://github.com/Microsoft/artifacts-credprovider/releases/download/v1.4.1/"
CREDENTIAL_PROVIDER_NET8 = CREDENTIAL_PROVIDER_BASE + "Microsoft.Net8.NuGet.CredentialProvider.tar.gz"
CREDENTIAL_PROVIDER_NET8_ZIP = CREDENTIAL_PROVIDER_BASE + "Microsoft.Net8.NuGet.CredentialProvider.zip"
CREDENTIAL_PROVIDER_NON_SC_VAR_NAME = "ARTIFACTS_CREDENTIAL_PROVIDER_NON_SC"
CREDENTIAL_PROVIDER_RID_VAR_NAME = "ARTIFACTS_CREDENTIAL_PROVIDER_RID"

def get_version(root):
    src = os.path.join(root, "src", "artifacts_keyring", "__init__.py")

    with open(src, "r", encoding="utf-8", errors="strict") as f:
        txt = f.read()

    m = re.search(r"__version__\s*=\s*['\"](.+?)['\"]", txt)
    return m.group(1) if m else "0.1.0"

def get_runtime_identifier():
    os_system = platform.system().lower()
    os_arch = platform.machine().lower()
    
    if os_system == "linux":
        runtime_id = "linux"
    elif os_system == "darwin":
        runtime_id = "osx"
    elif os_system == "windows":
        runtime_id = "win"
    else:
        print(f"Warning: Unsupported OS: {os_system}. Please set the {CREDENTIAL_PROVIDER_RID_VAR_NAME} environment variable to specify a runtime identifier.")
        return ""

    if "aarch64" in os_arch or "arm64" in os_arch:
        if (os_system == "windows"): # windows on ARM runs x64 binaries
            runtime_id += "-x64"
        else:
            runtime_id += "-arm64"
    elif "x86_64" in os_arch or "amd64" in os_arch:
        runtime_id += "-x64"
    else:
        print(f"Warning: Unsupported architecture: {os_arch}. Please set the {CREDENTIAL_PROVIDER_RID_VAR_NAME} environment variable to specify a runtime identifier.")
        return ""

    return runtime_id

def get_os_runtime_url(runtime_var):
    if runtime_var == "" and "osx" in runtime_var:
        return CREDENTIAL_PROVIDER_NET8_ZIP
    elif runtime_var == "":
        return CREDENTIAL_PROVIDER_NET8

    if "osx" in runtime_var:
        return CREDENTIAL_PROVIDER_NET8_ZIP.replace(".Net8", f".Net8.{runtime_var}")

    return CREDENTIAL_PROVIDER_NET8.replace(".Net8", f".Net8.{runtime_var}")

def get_download_url():
    # When building the platform wheels in CI, use the self-contained version of the credential provider.
    # In these cases, check ARTIFACTS_CREDENTIAL_PROVIDER_RID to determine the desired runtime identifier.
    if CREDENTIAL_PROVIDER_RID_VAR_NAME in os.environ and \
        os.environ[CREDENTIAL_PROVIDER_RID_VAR_NAME]:
            runtime_var = str(os.environ[CREDENTIAL_PROVIDER_RID_VAR_NAME]).lower()
            return get_os_runtime_url(runtime_var)
    
    # Specify whether they want self-contained auto-detection or not.
    # Only applicable for non-CI environments.
    use_non_sc = CREDENTIAL_PROVIDER_NON_SC_VAR_NAME in os.environ and \
        os.environ[CREDENTIAL_PROVIDER_NON_SC_VAR_NAME] and \
        str(os.environ[CREDENTIAL_PROVIDER_NON_SC_VAR_NAME]).lower() == "true"

    if use_non_sc:
        return CREDENTIAL_PROVIDER_NET8
    else:
        runtime_id = str(get_runtime_identifier())
        return get_os_runtime_url(runtime_id)


def download_credential_provider(dest):
    if not os.path.isdir(dest):
        os.makedirs(dest)

    print("Downloading and extracting artifacts-credprovider to", dest)
    download_url = get_download_url()
    print("Downloading artifacts-credprovider from", download_url)

    with urllib.request.urlopen(download_url) as download_file:
        if download_url.endswith(".zip"):
            with zipfile.ZipFile(io.BytesIO(download_file.read())) as zip_file:
                zip_file.extractall(dest)
        else:
            tar = tarfile.open(mode="r|gz", fileobj=download_file)

            # Python 3.12 adds a safety filter for tar extraction
            # to prevent placement of files outside the target directory.
            # https://docs.python.org/3.12/library/tarfile.html#tarfile.tar_filter
            if sys.version_info >= (3, 12):
                tar.extractall(dest, filter="data")
            else:
                tar.extractall(dest)

class BuildKeyring(build_py):
    def run(self):
        super().run()

class BuildKeyringPlatformWheel(bdist_wheel):
    def finalize_options(self):
        super().finalize_options()
        self.root_is_pure = False

class KeyringDistribution(Distribution):
    def has_ext_modules(self):
        return True

if __name__ == "__main__":
    root = os.path.dirname(os.path.abspath(__file__))
    dest = os.path.join(root, "src", "artifacts_keyring", "plugins")

    # Clean any previous build artifacts
    if os.path.exists(dest):
        print("Removing previous plugins artifacts in ", dest)
        shutil.rmtree(dest)

    download_credential_provider(dest)

	# Fix for liblttng-ust.so.0 not being found on Debian 12 and later.
    # See https://github.com/dotnet/runtime/issues/57784 for more info.
    clr_trace_path = os.path.join(
        dest,
        "plugins",
        "netcore",
        "CredentialProvider.Microsoft",
        "libcoreclrtraceptprovider.so",
    )
    if os.path.exists(clr_trace_path):
        print("Removing libcoreclrtraceptprovider.so from plugins directory")
        os.remove(clr_trace_path)
    
    setup(
        version=get_version(root),
        cmdclass={
            "build_py": BuildKeyring,
            "bdist_wheel": BuildKeyringPlatformWheel,
        },
        distclass=KeyringDistribution,
    )



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from __future__ import absolute_import

import json
import os
import platform
import requests
import subprocess
import sys

from . import __version__
from .support import Popen


class CredentialProvider(object):
    _NON_INTERACTIVE_VAR_NAME = "ARTIFACTS_KEYRING_NONINTERACTIVE_MODE"

    def __init__(self):
        if sys.platform.startswith("win"):
            tool_path = os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                "plugins",
                "plugins",
                "netcore",
                "CredentialProvider.Microsoft",
                "CredentialProvider.Microsoft.exe",
            )

            self.exe = [tool_path]
        else:
            tool_path_root = os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                "plugins",
                "plugins",
                "netcore",
                "CredentialProvider.Microsoft"
            )

            is_dotnet_runtime_required = False
            if os.path.exists(tool_path_root):
                # Ensure the plugins directory executable permissions are set so Python can execute
                # the Credential Provider plugin.
                try:
                    tool_path = os.path.join(tool_path_root, 'CredentialProvider.Microsoft')
                    if os.path.exists(tool_path):
                        os.chmod(tool_path, 0o755)
                except Exception as e:
                    raise RuntimeError(
                        "Failed to set executable permissions for the Credential Provider plugins directory "
                        + tool_path_root 
                        + ". Please ensure the directory has the correct access permissions (755). Error: "
                        + str(e)
                    )
                
                # If tool_path_root contains the runtimes directory, it means that the
                # binary is not self-contained and requires a .NET install to run.
                tool_path_runtimes = os.path.join(
                    tool_path_root,
                    "runtimes"
                )
                if os.path.exists(tool_path_runtimes):
                    is_dotnet_runtime_required = True

            if is_dotnet_runtime_required:
                tool_path = os.path.join(
                    tool_path_root,
                    "CredentialProvider.Microsoft.dll"
                )

                try:
                    # check to see if any dotnet runtimes are installed. Not checking specific versions.
                    output = subprocess.check_output(["dotnet", "--list-runtimes"]).decode().strip()
                    if(len(output) == 0):
                        raise Exception("No dotnet runtime found. Refer to https://learn.microsoft.com/dotnet/core/install/ for installation guidelines.")
                except Exception as e:
                    message = (
                        "Unable to find dependency dotnet, please manually install"
                        " the .NET runtime and ensure 'dotnet' is in your PATH. Error: "
                    )
                    raise Exception(message + str(e))
                
                self.exe = ["dotnet", "exec", tool_path]
            else:
                # for self-contained binaries, the executable is not the DLL
                if platform.system().lower() == "windows":
                    tool_path = os.path.join(
                        tool_path_root,
                        "CredentialProvider.Microsoft.exe"
                    )
                # linux and macOS
                else:
                    tool_path = os.path.join(
                        tool_path_root,
                        "CredentialProvider.Microsoft"
                    )
                
                self.exe = [tool_path]
            

        if not os.path.isfile(tool_path):
            raise RuntimeError("Unable to find credential provider in the expected path: " + tool_path)

    def get_credentials(self, url):
        # Public feed short circuit: return nothing if not getting credentials for the upload endpoint
        # (which always requires auth) and the endpoint is public (can authenticate without credentials).
        if not self._is_upload_endpoint(url) and self._can_authenticate(url, None):
            return None, None

        # Getting credentials with IsRetry=false; the credentials may come from the cache
        username, password = self._get_credentials_from_credential_provider(url, is_retry=False)

        # Do not attempt to validate if the credentials could not be obtained
        if username is None or password is None:
            return username, password

        # Make sure the credentials are still valid (i.e. not expired)
        if self._can_authenticate(url, (username, password)):
            return username, password

        # The cached credentials are expired; get fresh ones with IsRetry=true
        return self._get_credentials_from_credential_provider(url, is_retry=True)


    def _is_upload_endpoint(self, url):
        url = url[: -1] if url[-1] == "/" else url
        return url.endswith("pypi/upload")


    def _can_authenticate(self, url, auth):
        response = requests.get(url, auth=auth)

        return response.status_code < 500 and \
            response.status_code != 401 and \
            response.status_code != 403


    def _get_credentials_from_credential_provider(self, url, is_retry):
        non_interactive = self._NON_INTERACTIVE_VAR_NAME in os.environ and \
            os.environ[self._NON_INTERACTIVE_VAR_NAME] and \
            str(os.environ[self._NON_INTERACTIVE_VAR_NAME]).lower() == "true"

        proc = Popen(
            self.exe + [
                "-Uri", url,
                "-IsRetry", str(is_retry),
                "-NonInteractive", str(non_interactive),
                "-CanShowDialog", "True",
                "-OutputFormat", "Json"
            ],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )

        # Read all standard error first, which may either display
        # errors from the credential provider or instructions
        # from it for Device Flow authentication.
        for stderr_line in iter(proc.stderr.readline, b''):
            line = stderr_line.decode("utf-8", "ignore")
            sys.stderr.write(line)
            sys.stderr.flush()

        proc.wait()

        if proc.returncode != 0:
            stderr = proc.stderr.read().decode("utf-8", "ignore")

            error_msg = "Failed to get credentials: process with PID {pid} exited with code {code}".format(
                pid=proc.pid, code=proc.returncode
            )
            if stderr.strip():
                error_msg += "; additional error message: {error}".format(error=stderr)
            else:
                error_msg += "; no additional error message available, see Credential Provider logs above for details."
            raise RuntimeError(error_msg)

        try:
            # stdout is expected to be UTF-8 encoded JSON, so decoding errors are not ignored here.
            payload = proc.stdout.read().decode("utf-8")
        except ValueError:
            raise RuntimeError("Failed to get credentials: the Credential Provider's output could not be decoded using UTF-8.")

        try:
            parsed = json.loads(payload)
            return parsed["Username"], parsed["Password"]
        except ValueError:
            raise RuntimeError("Failed to get credentials: the Credential Provider's output could not be parsed as JSON.")
